#include <stdio.h>
#include "stm32f4xx.h"
#include "ws2812_stm32f4.h"
#include "bin_clock.h"

// NOTE: code does not have any safeguards against bad values for
// msb_index/nbits (-> buffer overflows), so choose wisely
typedef struct {
	uint32_t color_on;		// color/brightness when bit is set
	uint32_t color_off;		// ...
	uint32_t msb_index;		// index in "ws2812_framebuffer"
	uint32_t nbits;			// number of bits to display
} digit_def;

typedef struct {
	digit_def year;
	digit_def month;
	digit_def day;
	digit_def hour;
	digit_def minute;
	digit_def second;
	digit_def dayofweek;
	digit_def ampm; //am=on, pm=off
} disp_def;

#define BRT_ON		0x20
#define BRT_OFF		0x04

// color and brightness settings for the different values
/* LED arrangement for my display:
 * 47 ... 36
 * 35 ... 24
 * 23 ... 12
 * 11 ... 0
 */
static disp_def binary_display = {
		{ 0x00010000*BRT_ON, 0x00010000*BRT_OFF, 47, 12 }, 	//Y
		{ 0x00000100*BRT_ON, 0x00000100*BRT_OFF, 34,  4 },	//M
		{ 0x00000001*BRT_ON, 0x00000001*BRT_OFF, 23,  5 },	//D
		{ 0x00010100*BRT_ON, 0x00010100*BRT_OFF, 28,  5 },	//h
		{ 0x00000101*BRT_ON, 0x00000101*BRT_OFF, 17,  6 },	//m
		{ 0x00010001*BRT_ON, 0x00010001*BRT_OFF,  5,  6 },	//s
		{ 0x00010101*BRT_ON, 0x00010101*BRT_OFF,  9,  3 },	//dow
		{ 0x00010000*BRT_ON, 0x00000001*BRT_OFF, 29,  1 },	//ampm
};

// to remember if am/pm-indicator has to be displayed or not
static uint32_t Hour_Format = RTC_HourFormat_24;



// initialization of RTC
// h_format can be RTC_HourFormat_24 or RTC_HourFormat_12
int bin_clock_init(const uint32_t h_format)
{
	RTC_InitTypeDef rtcinit;
	ErrorStatus errcode; //0=error
	NVIC_InitTypeDef nvic_init;
	EXTI_InitTypeDef EXTI_InitStructure;

	if (IS_RTC_HOUR_FORMAT(h_format)) {
		Hour_Format = h_format;
	} else {
		Hour_Format = RTC_HourFormat_24;
	}

	RCC_APB1PeriphClockCmd(RCC_APB1Periph_PWR, ENABLE);

	PWR_BackupAccessCmd(ENABLE);
	RCC_LSICmd(ENABLE);
	while (RCC_GetFlagStatus(RCC_FLAG_LSIRDY) == RESET) {
	}

	RCC_RTCCLKConfig(RCC_RTCCLKSource_LSI);
	RCC_RTCCLKCmd(ENABLE);

	RTC_WaitForSynchro();

	RTC_WriteProtectionCmd(DISABLE);
	errcode = RTC_EnterInitMode();
	if (errcode == ERROR) {
		return -1;
	}

	// internal LSI frequency is ~32kHz
	// these should result in ~1Hz RTC clock
	// NOTE1: LSI is way to imprecise for time/calender purposes
	// use external OSC for real applications
	// NOTE2: LSE chrystal is not fitted on stm32f4discovery
	rtcinit.RTC_AsynchPrediv = 127;
	rtcinit.RTC_HourFormat = Hour_Format;
	rtcinit.RTC_SynchPrediv = 255;
	errcode = RTC_Init(&rtcinit);
	if (errcode == ERROR) {
		return -1;
	}

	errcode = RTC_WakeUpCmd(DISABLE);
	if (errcode == ERROR) {
		return -1;
	}

	// wakeup signal is connected to Line 22 (see manual)
	EXTI_ClearITPendingBit(EXTI_Line22);
	EXTI_InitStructure.EXTI_Line = EXTI_Line22;
	EXTI_InitStructure.EXTI_Mode = EXTI_Mode_Interrupt;
	EXTI_InitStructure.EXTI_Trigger = EXTI_Trigger_Rising;
	EXTI_InitStructure.EXTI_LineCmd = ENABLE;
	EXTI_Init(&EXTI_InitStructure);

	nvic_init.NVIC_IRQChannel = RTC_WKUP_IRQn;
	nvic_init.NVIC_IRQChannelPreemptionPriority = 0;
	nvic_init.NVIC_IRQChannelSubPriority = 0;
	nvic_init.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&nvic_init);

	// set wakeup counter to trigger every second
	RTC_SetWakeUpCounter(0x00);
	RTC_WakeUpClockConfig(RTC_WakeUpClock_CK_SPRE_16bits);
	RTC_ITConfig(RTC_IT_WUT,ENABLE);

	errcode = RTC_WakeUpCmd(ENABLE);
	if (errcode == ERROR) {
		return -1;
	}

	RTC_ExitInitMode();
	return 0;
}

// converts the value into an array of colors, one entry per bit
// on/off colors, size and destination pointer are defined in the "digit_def"-struct
static void convert(const digit_def * const digit, const uint32_t value) {
	uint32_t on = digit->color_on;
	uint32_t off = digit->color_off;
	volatile uint32_t * dest = &(ws2812_framebuffer[digit->msb_index]);
	uint32_t mask = 1 << (digit->nbits - 1); // if we need three bits the mask has to be 0x04 etc

	// check every significant bit in value and set framebuffer accordingly
	do {
		if (value & mask) {
			*(dest--) = on;
		} else {
			*(dest--) = off;
		}
		mask >>= 1;
	} while (mask != 0);
}

// gets the current RTC time and writes is into "ws2812_framebuffer"
// is automatically called from RTC_WKUP_IRQHandler
void display_time(void)
{
	RTC_TimeTypeDef time;
	RTC_DateTypeDef date;

	RTC_GetTime(RTC_Format_BIN, &time);
	RTC_GetDate(RTC_Format_BIN, &date);

	convert( &binary_display.year, 		date.RTC_Year+2000);
	convert( &binary_display.month, 	date.RTC_Month);
	convert( &binary_display.day, 		date.RTC_Date);
	convert( &binary_display.dayofweek, date.RTC_WeekDay);
	convert( &binary_display.hour, 		time.RTC_Hours);
	convert( &binary_display.minute, 	time.RTC_Minutes);
	convert( &binary_display.second, 	time.RTC_Seconds);

	if (Hour_Format == RTC_HourFormat_24) {
		ws2812_framebuffer[binary_display.ampm.msb_index] = 0;
	} else {
		if (time.RTC_H12 == RTC_H12_AM) {
			ws2812_framebuffer[binary_display.ampm.msb_index] =
					binary_display.ampm.color_on;
		} else {
			ws2812_framebuffer[binary_display.ampm.msb_index] =
					binary_display.ampm.color_off;
		}
	}
}

// set the RTC to a hard-coded default time
// does not write the time to the display
int set_default_time(void) {
	RTC_TimeTypeDef time;
	RTC_DateTypeDef date;
	ErrorStatus errcode; //0=error

	date.RTC_Year = 13;
	date.RTC_Month = 1;
	date.RTC_Date = 5;
	date.RTC_WeekDay = 6;
	time.RTC_H12 = RTC_H12_PM;
	time.RTC_Hours = 16;
	time.RTC_Minutes = 28;
	time.RTC_Seconds = 30;

	errcode = RTC_SetTime(RTC_Format_BIN, &time);
	if (errcode == ERROR) {
		return -1;
	}
	errcode = RTC_SetDate(RTC_Format_BIN, &date);
	if (errcode == ERROR) {
		return -1;
	}

	return 0;
}

// RTC wakeup irq, gets called once per second
void RTC_WKUP_IRQHandler(void)
{
	// reset IRQ source
	EXTI_ClearITPendingBit(EXTI_Line22);
	RTC_ClearITPendingBit(RTC_IT_WUT);

	// write current time to display buffer
	display_time();
}
