#include <stdio.h>
#include "stm32f4xx.h"
#include "ws2812_stm32f4.h"
#include "delay.h"

// externally writable color data for all leds
volatile uint32_t ws2812_framebuffer[WS2812_NR_LEDS];

// local buffer for timer/dma, one byte per bit + reset pulse
static uint16_t framebuffer[WS2812_FRAMEBUF_LEN];

// start the dma transfer (framebuffer to timer)
static void start_dma(void);




// writes the pwm values of one byte into the array which will be used by the dma
static inline void color2pwm(uint16_t ** const dest, const uint8_t color) {
	uint8_t mask = 0x80;

	do {
		if (color & mask) {
			**dest = WS2812_PWM_ONE;
		} else {
			**dest = WS2812_PWM_ZERO;
		}
		*dest += 1;
		mask >>= 1;
	} while (mask != 0);
}


// convert color codes from "ws2812_framebuffer" into serial "commands" for timer
void ws2812_update_leds(void)
{
	uint32_t i, tmp;
	uint16_t * bufp = framebuffer;

	for (i = 0; i < WS2812_NR_LEDS; i++) {
		tmp = ws2812_framebuffer[i];

		// edit here to change order of colors in "ws2812_framebuffer" (0x00RRGGBB, 0x00GGBBRR, etc)
		// the chip needs G R B
		color2pwm(&bufp, (tmp & 0x0000ff00) >> 8); // green
		color2pwm(&bufp, (tmp & 0x00ff0000) >> 16); // red
		color2pwm(&bufp, (tmp & 0x000000ff) >> 0); // blue
	}

	// start feeding the timer
	start_dma();
}

// initialization of peripherals for ws2812:
// GPIO, timer (PWM mode), DMA, NVIC
void ws2812_init(void)
{
	GPIO_InitTypeDef GPIO_InitStructure;
	TIM_TimeBaseInitTypeDef timbaseinit;
	TIM_OCInitTypeDef timocinit;
	NVIC_InitTypeDef nvic_init;

	// data init, make sure reset pulse is zero
	int i;
	for (i = 0; i < (WS2812_FRAMEBUF_LEN-WS2812_RESET_LEN); i++)
		framebuffer[i] = WS2812_PWM_ZERO;
	for (; i < WS2812_FRAMEBUF_LEN; i++)
		framebuffer[i] = 0;

	// GPIO
	RCC_AHB1PeriphClockCmd(WS2812_RCC_GPIO, ENABLE);
	GPIO_InitStructure.GPIO_Pin = WS2812_GPIO_PIN;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_100MHz;
	GPIO_Init(WS2812_GPIO, &GPIO_InitStructure);
	GPIO_PinAFConfig(WS2812_GPIO, GPIO_PinSource0, WS2812_GPIO_AF_TIM);

	// TIMER
	RCC_APB1PeriphClockCmd(WS2812_RCC_TIM, ENABLE);

	TIM_TimeBaseStructInit(&timbaseinit);
	timbaseinit.TIM_ClockDivision = TIM_CKD_DIV1;
	timbaseinit.TIM_CounterMode = TIM_CounterMode_Up;
	timbaseinit.TIM_Period = WS2812_TIM_FREQ / WS2812_OUT_FREQ;
	timbaseinit.TIM_Prescaler = (uint16_t)(
			(SystemCoreClock / 2) / WS2812_TIM_FREQ) - 1; //tim3 runs at half the speed of the system ->divide by 2
	TIM_TimeBaseInit(WS2812_TIM, &timbaseinit);

	TIM_OCStructInit(&timocinit);
	timocinit.TIM_OCMode = TIM_OCMode_PWM1;
	timocinit.TIM_OCPolarity = TIM_OCPolarity_High;
	timocinit.TIM_OutputState = TIM_OutputState_Enable;
	timocinit.TIM_Pulse = 0;
	WS2812_TIM_OC_INIT(WS2812_TIM, &timocinit);

	WS2812_TIM_OC_PRECFG(WS2812_TIM, TIM_OCPreload_Enable);
	TIM_ARRPreloadConfig(WS2812_TIM, ENABLE);

	TIM_CCxCmd(WS2812_TIM, WS2812_TIM_CHANNEL, TIM_CCx_Enable);
	TIM_Cmd(WS2812_TIM, ENABLE);

	// DMA
	RCC_AHB1PeriphClockCmd(WS2812_RCC_DMA, ENABLE);
	TIM_DMACmd(WS2812_TIM, WS2812_DMA_SOURCE, ENABLE);
	DMA_ITConfig(WS2812_DMA_STREAM, DMA_IT_TC, ENABLE);
	start_dma();

	// NVIC for DMA
	nvic_init.NVIC_IRQChannel = DMA1_Stream7_IRQn;
	nvic_init.NVIC_IRQChannelPreemptionPriority = 4;
	nvic_init.NVIC_IRQChannelSubPriority = 0;
	nvic_init.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&nvic_init);
}


// begin transferring framebuffer data to the timer
static void start_dma(void)
{
	static DMA_InitTypeDef dma_init =
	{
			.DMA_BufferSize 		= WS2812_FRAMEBUF_LEN,
			.DMA_Channel 			= WS2812_DMA_CHANNEL,
			.DMA_DIR 				= DMA_DIR_MemoryToPeripheral,
			.DMA_FIFOMode 			= DMA_FIFOMode_Disable,
			.DMA_FIFOThreshold 		= DMA_FIFOThreshold_HalfFull,
			.DMA_Memory0BaseAddr 	= (uint32_t) framebuffer,
			.DMA_MemoryBurst 		= DMA_MemoryBurst_Single,
			.DMA_MemoryDataSize 	= DMA_MemoryDataSize_HalfWord,
			.DMA_MemoryInc 			= DMA_MemoryInc_Enable,
			.DMA_Mode 				= DMA_Mode_Normal,
			.DMA_PeripheralBaseAddr = (uint32_t) &WS2812_TIM->WS2812_CCR,
			.DMA_PeripheralBurst 	= DMA_PeripheralBurst_Single,
			.DMA_PeripheralDataSize = DMA_PeripheralDataSize_HalfWord,
			.DMA_PeripheralInc 		= DMA_PeripheralInc_Disable,
			.DMA_Priority 			= DMA_Priority_Medium
	};

	DMA_Cmd(WS2812_DMA_STREAM, DISABLE);
	while (DMA_GetCmdStatus(WS2812_DMA_STREAM) == ENABLE)
		; //wait
	DMA_Init(WS2812_DMA_STREAM, &dma_init);
	DMA_Cmd(WS2812_DMA_STREAM, ENABLE);
	TIM_DMACmd(WS2812_TIM, WS2812_DMA_SOURCE, ENABLE);
}


// gets called when dma transfer has completed
void DMA1_Stream7_IRQHandler(void)
{
	DMA_ClearITPendingBit(WS2812_DMA_STREAM, DMA_IT_TCIF7);
	DMA_Cmd(WS2812_DMA_STREAM, DISABLE);

	// need to disable this, otherwise some glitches can occur (first bit gets lost)
	TIM_DMACmd(WS2812_TIM, WS2812_DMA_SOURCE, DISABLE);
}


#define NR_TEST_PATTERNS	8
void ws2812_test_pattern(void)
{
	uint32_t i,j;
	uint32_t patterns[NR_TEST_PATTERNS] = {
			0x000f0000,
			0x00000f00,
			0x0000000f,
			0x000f0f00,
			0x000f000f,
			0x00000f0f,
			0x00ffffff,
			0x00000000,
	};

	for (i = 0; i < WS2812_NR_LEDS; i++) {
		ws2812_framebuffer[i] = 0;
	}

	for(j=0; j<NR_TEST_PATTERNS; j++)
	{
		for (i = 0; i < WS2812_NR_LEDS; i++)
		{
			ws2812_framebuffer[i] = patterns[j];
			ws2812_update_leds();
			delay_ms(10);
		}
	}

	for (i = 0; i < WS2812_NR_LEDS; i++) {
		ws2812_framebuffer[i] = 0;
	}
}
